/*
  This program is free software published under the
  terms of the GNU Lesser General Public License.
  http://www.gnu.org/copyleft/lesser.html    
  This program is distributed in the hope that it will
  be useful - WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A
  PARTICULAR PURPOSE.
*/ 


/*  xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx
     source code: ES6 (ECMAScript 6, ECMAScript 2015, JavaScript 6)
    xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx */

//  xxxxxxxxxxxxx          version: 1.0              xxxxxxxxxxxxxxx
		 
//  xxxxxxxxxxxxxxx   global variable declarations   xxxxxxxxxxxxxxxx

 
// These variables below can be directly modified by the user on the options form: 

var variable = '2';  // Variable selector: '1' or '2' or '3'  

var iter = 1;   // Number of iterations for each variable value

var s_time = [];     //   Variable array. Alphanumeric characters duration (ms) => Duration of main string stimulus (digits or letters)

var d_time = [];     // Variable array. Marks (cues) delay time (ms)

var m_time = [];  		// variable array. Marks duration time (ms)

var before = 1500;       // const. Delay time before show each stimulus (ms).

var after = 1000;				 // const. Delay time after show each stimulus (ms)

var alfa_type = '3';  // Type of Alphanumeric characters => "1" = Digits; "2" = Letters-lowercase; "3" = Letters-UPPERCASE; "4" = All digits and letters

var target_cnt = 1;   // Number of characters to memorize. range: [1..row_cnt * col_cnt]

var row_cnt = 3;   // Number of rows in each item

var col_cnt = 3;  // Number of columns in each item. Number of of alphanumeric characters in each row. 

var mtype = '2';       // Type of marking. range: '1'..'4'

var font_family = 'Georgia, serif';  // default canvas ctx.fontFamily 

var is_back = false;    // if (is_back==true) => user can go back to previous item => user can only change answer to prior item. Each item id displayed only once.  

var is_mis = true;    // if (is_mis==true) => missing values are allowed

var prg_type = true;   //  type of information in Progress Bar (true => Items, false => Nothing) 

var demo = true;     // if (demo == true) => Show task solutions 

	// var limit = true;  // Set user answer textbox maximum length to Number of characters to memorize: true or false
  
var back_clr = '#fff';  // background canvas color during task

var pen_clr = '#000';  //  font color (base font) during the task 
	
var store = false;      // false => this web application will not store results within client web browser.; true => This web application will store the results inside client web browser. 
 
// Other global constants and variables:

var item_cnt;     // total number of items in the task

var item = 0;   // current item in task  

var is_storage = false; // local storage is enabled on user browser or not
   
var dirty = false;  // if  (dirty==true) => user changed any options in form 

var is_task = false;    // task in progress or not. is_task is true for all events between startTask() and theEnd() events

var startTime;      // timestamp (start time) for current stimulus

var counter = 0;     // toogle counter (0 or 1) for window.requestAnimationFrame(prepare) callback and window.requestAnimationFrame(getFPS) callback

var FPS = 0;     // frames per second. Measured Frame rate on user screen. see: function getFPS() 

// global (window) arrays
	
var stim = [];   // stimulus string content for each item. dimension range: [0..item_cnt-1]

var ans = [];   //  the user's response to each item in the task

var target = [];     //  For each item in the task: which character is the target, ie which character should be remembered 

var repeated = [];  // for each item: the number of completed repetitions of the stimulus for each item. see: showItem() 


// application init [check storage, show start panel..] 
function init() 
{   
  is_storage= isStorage() && storageAvailable('localStorage');
  if (is_storage)  {
    document.getElementById('store_group').style.display= 'block';
		document.getElementById('no_storage').style.display= 'none';
  }			
	let e= document.getElementById('answer');  // attach event handlers..
	e.onkeypress = answerKeyPress;	
	e.onkeyup = answerKeyUp;	
	e.onpaste= function(){return false}; // Prevent pasting in answer input type=text (since pasted content might include non-allowed characters)

	// do FPS calibration..	
	setCanvas(document.getElementById('can_s'));	
	window.requestAnimationFrame(getFPS);  

	setDefault();
	loadFont();
	setIns();
}


// set default values in all <textarea> objects => variables
function setDefault()
{ 
	document.getElementById('s_time').value= '10\n100\n500';
	document.getElementById('d_time').value= '100\n500\n200';
	document.getElementById('m_time').value= '100\n1000\n500';
}


/* https://www.viewsonic.com/library/tech/frame-rate-vs-refresh-rate-whats-the-difference/
	 https://en.wikipedia.org/wiki/Frame_rate
	Frame rate is produced by a combination of your graphics card and your processor, so it is essentially the number of frames a system is able to produce in a second, whereas refresh rate is the rate at which a monitor is able to completely refresh its display.  */
function getFPS(timestamp) 
{ 
	if (counter == 0) {  
		startTime = timestamp;	// Remember the time when the stimulus display started 
		//FPS= 0;  // redundant
		document.getElementById('can_s').style.display= 'block';  	  	
	}	
	counter= 1;	
	let elapsed = timestamp - startTime;  // const elapsed = Math.round(timestamp - start); 						
	if ( elapsed <= 1000  ) {    // just wait
		FPS++;
		window.requestAnimationFrame(getFPS);  	// Note: Your callback routine must itself call requestAnimationFrame() again if you want to animate another frame at the next repaint. requestAnimationFrame() is 1 shot.				
	}	
	else  {  // hide stimulus 	
	  let freq= (elapsed/FPS).toFixed(3);
	  document.getElementById('fps').textContent= ' Measured Frame rate (frames per second or FPS or frame frequency) FPS = ' + FPS + ' Hz. Tested time period = 1000 ms.';
    document.getElementById('fps_2').textContent=  'Theoretical Minimum duration of stimuli on this device = ' + freq + 'ms. Maximum reliable step value to increase stimulus length on this device = ' + Math.round(freq) + 'ms. Do not set lower values for the step (increase or decrease) of any variable than this value. Each value of each variable must be greater than this value (' + Math.round(freq) + 'ms).';		
		document.getElementById('can_s').style.display= 'none';	
		showStart();   		
		//alert('Calibration done. \n Measured Frame rate (FPS, frames per second, frame frequency) FPS = ' + FPS + ' Hz. \n Tested time period = 1000 ms. \n  Theoretical Minimum duration of stimuli on this device = ' + freq + ' ms. \n  Maximum reliable step value to increase stimulus length on this device = ' + freq + ' ms.');
	}			
}


// fill font-family names from array into select html element
function loadFont()  
{
	// used font-family array: The name of a font family. For example, "Times" and "Helvetica" are font families. Font family names containing whitespace should be quoted
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp	
	const arr = ['"Times New Roman", Times, serif', 'Georgia, serif', '"Palatino Linotype", "Book Antiqua", Palatino, serif', 'Arial, Helvetica, sans-serif', '"Arial Black", Gadget, sans-serif', '"Comic Sans MS", cursive, sans-serif', 'Impact, Charcoal, sans-serif', '"Lucida Sans Unicode", "Lucida Grande", sans-serif', 'Tahoma, Geneva, sans-serif', '"Trebuchet MS", Helvetica, sans-serif', 'Verdana, Geneva, sans-serif', '"Courier New", Courier, monospace', '"Lucida Console", Monaco, monospace' ];
	let sel= document.getElementById('font_family'), len = arr.length, opt; 
	sel.innerHtML= ''; 
	for (let i= 0; i!= len; i++) 
	{  
		opt = document.createElement('option');
		opt.value= arr[i];  // option value selected font-family name (string)            
		opt.text= arr[i];   // option textContent => font-family name (string)              		                       								
		sel.add(opt);  // sel.add(opt, null);                 			
	} 
	sel.selectedIndex= 1; // Georgia, serif
	document.getElementById('font_family_msg').style.fontFamily= sel.value; // = arr[1];
}	


// before using web storage, check browser support for localStorage and sessionStorage
function isStorage()  
{
  if (typeof(Storage) !== 'undefined') {
     return true;                 
  } 
  else {
    return false;    
  }  
}


// necessarily because of IE11 offline mode
function storageAvailable(type) 
{
  try  {
    return (window[type]) ? true: false;   
  }
  catch(e) {
    return false;
  }
}


// variable => basic textarea validator => onkeydown() => see html tag
function updateVar(event)
{
	event = event || window.event;
	let key = event.keyCode || event.which;
	if( key == 13 || key == 37 || key == 38 || key == 39 || key == 40 || key == 8 || key == 46 ) { // ENTER, Left / Up / Right / Down Arrow, Backspace, Delete keys
		return;  // Don't validate the input if ENTER, arrows, delete and backspace keys were pressed 
	}
	let is_digit = (key >= 48 && key <= 57) || (key >= 96 && key <= 105);
	if( !is_digit )  {
	event.returnValue = false;
	if(event.preventDefault) 
		event.preventDefault();
	}   
}


// show start panel 
function showStart()  
{      
  document.getElementById('end').style.display= 'none';
  document.getElementById('result_wrapper').style.display= 'none';  
  document.getElementById('store_id').style.display= 'none';
	document.getElementById('table_grp').style.display= 'none';
	document.getElementById('start').style.display= 'block';	
}


// raised when user change font-family on html <select> element
function tryFont(e)  {	
	document.getElementById('font_family_msg').style.fontFamily= e.value;
}	


// set instructions text depending on user selected task options
function setIns()  
{
  document.getElementById('instructions').style.display= 'block';		   	
  document.getElementById('ins_back').style.display = is_back ? 'block' : 'none';  
	document.getElementById('ins_demo').style.display = demo ? 'block' : 'none';	
	let e= document.getElementById('mtype');  
	if ( e.selectedIndex > - 1 )  {		
		document.getElementById('ins_mtype').textContent= '(' + e.options[e.selectedIndex].textContent + ')';
	}
}	


// raised when page is resized. see html: <body onresize="pageResize()"> => stretch canvas 
function pageResize()  
{   
	if (is_task)  { 
		theEnd(true);	  // resize error										 
	}	
}	


// refresh Options (update *_msg label) => give user short option HELP
function refOpt(e)  
{   
  dirty=true;       
  let my_id = e.id, msg = document.getElementById(my_id + '_msg'), s = 'Values within this range are allowed: ', s2= ' with step= ';               
	switch (my_id)
	{

    case 'variable': 
			msg.textContent= e.options[e.selectedIndex].getAttribute('data-help');	
		  for (let i= 1; i<=3; i++)  {
				document.getElementById('fs_' + i).style.backgroundColor= ( e.value == i ) ? '#e5e5e5': '#fff'; 
			} 
		  break;

    case 'iter':	 
		case 'after':		
			if (e.validity.valid)  {
				msg.textContent= ''; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 			
			break;  
		
		case 'before':
			if (e.validity.valid)  {
				msg.textContent= 'Stimuli are prepared during this period. So leave this period longer, especially for very short stimulus displays.'; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 			
      break;

		case 'row_cnt':		
    case 'col_cnt': 	
		case 'target_cnt':			
			if (e.validity.valid)  {
				msg.textContent= ''; 
				msg.removeAttribute('class');
			}
			else  {
				msg.textContent= s + e.min + ".." + e.max + s2 + e.step;
				msg.setAttribute('class', 'error');
			} 
      let row = parseInt( document.getElementById('row_cnt').value, 10);			
			let col = parseInt( document.getElementById('col_cnt').value, 10);
			let tar_cnt = parseInt( document.getElementById('target_cnt').value, 10);
			let val= document.getElementById('target_cnt_msg');
			if ( tar_cnt > row * col ) {
				val.textContent= 'Number of characters to memorize is to big';
				val.setAttribute('class', 'error');
			}
			else { 
			  val.textContent= ''; 
				val.removeAttribute('class');
			}
			break;
		
		/*case 'font_family':
		  break;
    
		case 'mtype':
      break; 		

		case 'alfa_type':	 
			break;  	*/		
		
    case 'is_back':  		
			msg.textContent= e.checked ? 'Yes. You can go back to the previous item and correct the previously entered values. However, you cannot re-display previously displayed alphanumericals.' : 'No. Returning to previous items is not allowed.';    
			break;
    
    case 'is_mis': 
    	msg.textContent= e.checked ? 'Missing values are allowed. The user does not have to enter a value. For next item: press ENTER key or "Right Arrow" button' : 'Missing values are not allowed. The user must enter any non-empty value.';    
			break;
			    
    case 'prg_type':                                                                                                                                                   
      msg.textContent= e.checked ? 'Progress bar will show the number of the current item and the number of remaining items until the end of the task.' : 'Progress bar will be invisible';     
      break;    
				
		case 'demo':
			msg.textContent= e.checked ? 'The correct solution will be displayed for each item.' : 'No';  
			break; 
		
	  case 'limit':
			msg.textContent= e.checked ? 'Yes' : 'No';  
			break; 
		
		case 'back_clr': 		// background canvas color during task
		case 'pen_clr':  		//  font color (base font) during the task   
			msg.textContent= ( validColor(e.value) ) ? '' : 'The selected color is invalid.'; 
			break; 	
		
		case 'store':                                                                                                                                           
			msg.textContent= (e.value === '1') ? 'This web application will store the results inside your web browser (no expiration date). Useful for group research' : 'This web application will not store results within your web browser. Data is lost when the browser tab is closed';
			break;
 
	}   // switch
	
  refMsg();   
}


// refresh main validator message for user defined OPTIONS 
function refMsg()   
{ 	
	if ( optionsValid() )  {
    document.getElementById('is_valid').textContent=  dirty == true ? 'All your settings have been successfully updated' : '';  
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#000000'); 
  }
  else  {
    document.getElementById('is_valid').textContent='Some of your options are incorrectly set. They were not updated.';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold'); 
  } 	
}


// only check for validity.valid object in all input type="number" + check colors
function optionsValid()  
{      
	let res= document.getElementById('iter').validity.valid &&
		document.getElementById('before').validity.valid &&	
		document.getElementById('after').validity.valid &&	
	  document.getElementById('target_cnt').validity.valid &&
	  document.getElementById('row_cnt').validity.valid && 	
		document.getElementById('col_cnt').validity.valid && 	
		validColor(document.getElementById('back_clr').value) && 
		validColor(document.getElementById('pen_clr').value);
  return res;              
}


// start task (research) 
function startTask(e)   
{     
  initLayout();  

	setPar(); 
	
	stimGen();
 
  showItem();     
}


// init page layout when task is started. init visible/invisible html dom elements, progress bar
function initLayout()
{

	setCanvas(document.getElementById('can_s'));	
	setCanvas(document.getElementById('can_m'));	
	
	document.getElementById('result_wrapper').style.display= 'none';
  document.getElementById('end').style.display= 'none'; 
  document.getElementById('start').style.display= 'none';              
  document.getElementById('task_section').style.display= 'block'; 
	document.getElementById('item_div').style.display= 'none';
	
	document.getElementById('nav_items').textContent= '';	
  if (!prg_type)  {
		document.getElementById('progress_holder').style.display= 'none';    	
	}	
	else  {
		document.getElementById('progress_holder').style.display= 'block';   				   			
		document.getElementById('progress').setAttribute('style', 'width:100%');  // reset progress bar width						 	
	}
  
	document.getElementById('demo_sol').style.display= demo ? 'block': 'none';
	
	switch (alfa_type)
	{		
		case '2':  // Letters-lowercase 
			document.getElementById('answer').setAttribute('style','text-transform:lowercase');			
			break;
		case '3':  // Letters-UPPERCASE 
			document.getElementById('answer').setAttribute('style','text-transform:uppercase');			
			break;
		default:  // Digits ("1") or All digits and letters ("4")
			document.getElementById('answer').removeAttribute('style');	
	} 
	
	if ( document.getElementById('limit').checked )  {
		document.getElementById('answer').setAttribute('maxLength', target_cnt);				
	}
	else  {
		document.getElementById('answer').setAttribute('maxLength', 100);
	}
}


// set stimulus canvas on browser screen
// c => canvas element => document.getElementById('can_s') or document.getElementById('can_m')
function setCanvas(c)  
{
	let ctx= c.getContext('2d');	 
	c.width= innerWidth;  
	c.style.width= c.width + 'px';	     
	c.height= innerHeight;   
	c.style.height= c.height + 'px';  

  ctx.clearRect(0, 0, c.width, c.height);  
	c.style.display= 'block';
	
	ctx.textAlign = 'center';
	ctx.textBaseline= 'middle';	
	
	ctx.fillStyle= back_clr;
	ctx.strokeStyle= pen_clr; 	
	
	// https://www.w3schools.com/cssref/css_websafe_fonts.asp
	// ctx.font= '10px Arial, Verdana, Helvetica, sans-serif';  	
}


// set internal application research parameters
// reset all global vars and arrays
function setPar()  
{
	
	if ( target_cnt > row_cnt * col_cnt ) {
    target_cnt= row_cnt * col_cnt;  // maximum target cnt
		document.getElementById('target_cnt').value= target_cnt;
	}		
	
	//validate();
		
	s_time = [];     // window variable array. Alphanumeric characters duration (ms) => Duration of main stimulus (digits or letters)
  d_time = [];     // global window variable array. Marks (cues) delay time (ms)
	m_time = [];  	 // global variable array. Marks duration time (ms)
	
	let lines, v, default_1, default_2;	
	switch (variable) 
	{
		case '1': 	// set s_time => variable = '1'
		  lines = document.getElementById('s_time').value.split('\n');
			for (let i = 0; i < lines.length; i++)  
			{
				v = parseInt(lines[i], 10);  
				for (let j= 1; j<= iter; j++) {   // each valid variable value x iter (number of iterations)
					s_time.push(v);
				}
			}	
			if  ( s_time.length > 1 ) {
				s_time= shuffle(s_time);
			}				
		  item_cnt= s_time.length;   // xxxxxx
			default_1= parseInt(document.getElementById('d_time').value.split('\n')[0], 10);  // only 1. valid value is used (1. row in textarea)
			default_2= parseInt(document.getElementById('m_time').value.split('\n')[0], 10);  // only 1. valid value is used (1. row in textarea)
			for (let i=0; i < item_cnt; i++)  {
				d_time[i] = default_1;
        m_time[i] = default_2;				
			}			
		  break;
		
		case '2':   // set d_time => variable = '2' 
		  lines = document.getElementById('d_time').value.split('\n');
			for (let i = 0; i < lines.length; i++)  {
				v = parseInt(lines[i], 10);  
				for (let j= 1; j<= iter; j++) {   // each valid variable value x iter (number of iterations)
					d_time.push(v);
				}
			}	
			if  ( d_time.length > 1 ) {
				d_time= shuffle(d_time);
			} 
		  item_cnt= d_time.length;   // xxxxxx
			default_1= parseInt(document.getElementById('s_time').value.split('\n')[0], 10);  
			default_2= parseInt(document.getElementById('m_time').value.split('\n')[0], 10);  
			for (let i=0; i < item_cnt; i++)  {
				s_time[i] = default_1;
        m_time[i] = default_2;				
			}			
		  break;
			
		case '3': 	// set m_time => variable = '3'
			lines = document.getElementById('m_time').value.split('\n');
			for (let i = 0; i < lines.length; i++)  {
				v = parseInt(lines[i], 10);  
				for (let j= 1; j<= iter; j++) {   // each valid variable value x iter (number of iterations)
					m_time.push(v);
				}
			}	
			if  ( m_time.length > 1 ) {
				m_time= shuffle(m_time);
			}		
		  item_cnt= m_time.length;    // xxxxxx
			default_1= parseInt(document.getElementById('s_time').value.split('\n')[0], 10);  
			default_2= parseInt(document.getElementById('d_time').value.split('\n')[0], 10);  
			for (let i=0; i < item_cnt; i++)  {
				s_time[i] = default_1;
        d_time[i] = default_2;				
			}			
		  break;				
	}
	
	// init global arrays
	stim = [];
	ans = []; 	
	target = []; 
	repeated = [];
	// after setting item_cnt
	for (let i= 0; i< item_cnt; i++)   {  // i => item        
    ans[i]= '';	
		repeated[i]= 0;
  } 
	
	// reset all global vars
  item= 0;  	
	is_task= true;     //  see pageResize()
	counter = 0;		
}


// validates and filter variable values entered in all textarea elements. set default value
function validate()
{		
	let s_time= [], d_time= [], m_time= [], lines, v, sel = {}, res; // Note: 3 arrays have local scope
	
	// s_time variable = '1'. range= [10..5000]. default= 100
  lines = document.getElementById('s_time').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);  // code here using lines[i] which will give you each line
		if ( isInt(lines[i]) )  { 
		  v = parseInt(lines[i], 10);
			if ( v >= 10 && v <= 5000 && typeof sel[v] === 'undefined')  {
				s_time.push(v);
				sel[v]= 1;
			}
		}
	}
	if (s_time.length == 0)  {
		s_time.push(100);
	}	 	
	document.getElementById('s_time').value= variable == '1' ? s_time.join('\n') : s_time[0];  // must be only one value if this var is not selected. see: setPar()
		
	// d_time variable = '2'. range=[0..5000]. default value = 100.
	sel = {};
	lines = document.getElementById('d_time').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);  
		if ( isInt(lines[i]) )  { 
		  v = parseInt(lines[i], 10);
			if ( v >= 0  &&  v <= 5000 && typeof sel[v] === 'undefined' )  {
				d_time.push(v);
				sel[v]= 1;
			}
		}
	}
	if (d_time.length == 0)  {
		d_time.push(100);
	}	 	
	document.getElementById('d_time').value= variable == '2' ? d_time.join('\n') : d_time[0]; // must be only one value if this var is not selected. see: setPar()
		
	// m_time variable = '3'. range= [10..5000]. default: 100
	sel = {};
	lines = document.getElementById('m_time').value.split('\n');
	for (let i = 0; i < lines.length; i++)
	{
		lines[i]= trim(lines[i]);
		if ( isInt( lines[i]) )  {  
		  v = parseInt(lines[i], 10);
			if ( v >= 10  &&  v <= 5000 && typeof sel[v] === 'undefined' )  {
				m_time.push(v);
				sel[v]= 1;
			}
		}
	}
	if (m_time.length == 0)  {
		m_time.push(100);
	}	
	document.getElementById('m_time').value= variable == '3' ? m_time.join('\n') : m_time[0];  // must be only one value if this var is not selected. see: setPar()
	
	switch (document.getElementById('variable').value) 
	{
		case '1': 
		  res = s_time.length > 1 ? true : false;  
		  break;
			
		case '2': 
		  res = d_time.length > 1 ? true : false;  
		  break;

		case '3': 
		  res = m_time.length > 1 ? true : false;  
		  break;			
	}
	
	if (!res)  {
    document.getElementById('is_valid').textContent+='Variable used as constant (only one valid value).';   // These settings are marked in red. 
    document.getElementById('is_valid').setAttribute('style', 'text-align:center;color:#ff0000;font-weight:bold');   	
	}
}	


// isNaN("") = false: the empty string is converted to 0 which is not NaN                   
function isInt(n) {       
   return (!isNaN(n)) && (parseFloat(n) === parseInt(n, 10)) &&  (n % 1 === 0); 
} 


// unbiased shuffle algorithm: Fisher-Yates (aka Knuth) Shuffle.
function shuffle(arr) 
{
  let cur_index = arr.length, tmp, rnd_index;
  // While there remain elements to shuffle...
  while (0 !== cur_index) {    
    rnd_index = Math.floor(Math.random() * cur_index);     // Pick a remaining element...
    cur_index = cur_index - 1;   
    tmp = arr[cur_index];          // And swap it with the current element.
    arr[cur_index] = arr[rnd_index];
    arr[rnd_index] = tmp;
  }
  return arr;
}


// stimulus generator. generates task stimuli 
function stimGen()	
{
	// generate string content	
	//stim = []; => see setPar() 
	for (let item= 0; item < item_cnt; item++)  {		
		stim[item]= genItem(item);
	}
	
	// determine which character in each item/stimulus should be memorized 
	let sel, ind, arr=[], max= row_cnt * col_cnt -1;   
	for (let item= 0; item < item_cnt; item++)   
	{
    target[item]= [];
		sel= {};
		arr= [];  // tmp array for all index within item sorting ascending
		for (let j= 0; j< target_cnt; j++)  
		{
			do  {
			  ind = getRndInt(0, max);   // or getRndInt(0, stim[item].length-1);
			} while (sel[ind] == '1');
			sel[ind]= 1;
			arr[j]= ind;  // index of target character in this item
		}
		arr.sort(function(a, b) {
			return a - b;
		});
		target[item]= arr;  // must be sorted
  }		
}


/* generate new item. i => item number. range= [0..item_cnt-1]
   each character in each row must be different */ 
function genItem(i) 
{
	let result= ''; 	
	for (let row = 0; row < row_cnt; row++)
	{
		let s= '';  // characters in each row 
		switch (alfa_type)
		{
			case '1':  // digits 	=> duplicates are always allowed (within each row and col)
				let d;  // let d, key= {};
				for (let i=0; i < col_cnt; i++)  {
						//do  {
					d= getRndInt(0,9).toString();
						//}	while ( key[d] == 1 );   // each character in each row must be different
						//key[d] = 1;
					s+= d;
				}			  
				break;			
			case '2':  // Letters-lowercase => no duplicates only within each row
				s= lowerStrGen(col_cnt);	// html element. col_cnt.max= 10		
				break;
			case '3':  // Letters-UPPERCASE => no duplicates only within each row
				s= upperStrGen(col_cnt);			
				break;
			default:  // 4 => All digits and letters => no duplicates only within each row
				s= allStrGen(col_cnt);
		}					
		result+= s;
	}
	return result;
}


// generates a random string from lowercase letters
// each generated character must be different
function lowerStrGen(cnt)
{ 
  let all = "abcdefghijklmnopqrstuvwxyz";  // length= 26
  let txt = "", up= all.length, key= {}, c;
  for (let i = 0; i < cnt; i++)  {
		do  {
			c= all.charAt(Math.floor(Math.random() * up));
		}	while ( key[c] == 1 ); 
		key[c]= 1;
    txt += c;
	}
  return txt;
}


// generates a random string from uppercase letters
// each generated character must be different 
function upperStrGen(cnt)
{ 
  let all = "ABCDEFGHIJKLMNOPQRSTUVWXYZ";   // length= 26
  let txt = "", up= all.length, key= {}, c;
  for (let i = 0; i < cnt; i++)  {
    do  {
			c= all.charAt(Math.floor(Math.random() * up));
		}	while ( key[c] == 1 ); 
		key[c]= 1;
		txt += c;
	}
  return txt;
}


// generates a random string from all (uppercase and lowercase) letters and digits
// each generated character must be different 
function allStrGen(cnt)
{ 
  let all = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789";
  let txt = "", up= all.length, key= {}, c;
  for (let i = 0; i < cnt; i++)  {
    do  {
			c= all.charAt(Math.floor(Math.random() * up));
		}	while ( key[c] == 1 ); 
		key[c]= 1;
		txt += c;
	}
  return txt;
}


// Changes application settings that can be changed directly by the user within the options form
function setVarPar()  
{   
  
	if (document.getElementById('variable').selectedIndex > -1 )  {
		variable = document.getElementById('variable').value;
  }
	
  if ( document.getElementById('iter').validity.valid )  {  
		iter= parseInt(document.getElementById('iter').value, 10);			
  }

  if (document.getElementById('before').validity.valid) {
		before= parseInt(document.getElementById('before').value, 10);
  }
	  
	if (document.getElementById('after').validity.valid) {
		after= parseInt(document.getElementById('after').value, 10);  
  }

  if ( document.getElementById('alfa_type').selectedIndex > - 1 )  {
		alfa_type= document.getElementById('alfa_type').value;
	}

  if ( document.getElementById('target_cnt').validity.valid )  {  
		target_cnt= parseInt(document.getElementById('target_cnt').value, 10);			
  }

  if ( document.getElementById('row_cnt').validity.valid )  {  
		row_cnt= parseInt(document.getElementById('row_cnt').value, 10);			
  } 

  if ( document.getElementById('col_cnt').validity.valid )  {  
		col_cnt= parseInt(document.getElementById('col_cnt').value, 10);			
  } 

  if (document.getElementById('mtype').selectedIndex > -1 )  {
		mtype = document.getElementById('mtype').value;
  }

	font_family= document.getElementById('font_family').value;

	is_back= document.getElementById('is_back').checked ? true: false;       
  
  is_mis= document.getElementById('is_mis').checked ? true: false; 
	
	prg_type= document.getElementById('prg_type').checked ? true : false;  // display item count on footer or nothing
		
	demo= document.getElementById('demo').checked ? true: false;	
		
	if (validColor(document.getElementById('back_clr').value))  {		
		back_clr= document.getElementById('back_clr').value;
  }
	
	if (validColor(document.getElementById('pen_clr').value))  {		
		pen_clr= document.getElementById('pen_clr').value;
  }
		
  store= (document.getElementById('store_group').disabled === false && document.getElementById('store').value === '1') ? true : false; 
}


// test color for validity (from string)
// https://stackoverflow.com/questions/8027423/how-to-check-if-a-string-is-a-valid-hex-color-representation/8027444
function validColor(clr){
  return /^#[0-9A-F]{6}$/i.test(clr);
}


// goto previous item (stimulus)
function priorItem()  
{  
  ans[item]= trim(document.getElementById('answer').value);
	if (item > 0)  {
    item--;  
    showItem();     // show again Panel for previous item 
  }         
}


// goto next item (stimulus)
function nextItem()  
{ 	
	if ( !valAnswer() )  {  
		document.getElementById('req_val').style.display= 'block';    //alert('You must enter any non-blank answer. Try to guess.'); 
		return;
  }	  
	ans[item]= trim(document.getElementById('answer').value);	
	item++;					
	if (item < item_cnt)  {         		 
    showItem();     // show next item        
  }
  else  {               
    theEnd(false);  // end of task. no resize error in this task 
  }
}


// trim input string (user answer)
function trim(str) {
   return str.replace(/^\s\s*/, '').replace(/\s\s*$/, '');
}
 
 
/* see function init().  deprecated but still works:
   https://developer.mozilla.org/en-US/docs/Web/API/GlobalEventHandlers/onkeypress
   this regex [1-9()*\/+-] was tested here:    https://regex101.com/    */
function answerKeyPress(event)   { 
  if (alfa_type == '1') {   // only digits
		return event.charCode === 0 || /[0-9]/.test(String.fromCharCode(event.charCode));	
	}
	else {
		return true;
	}
}


// go to the next task if the ENTER key is pressed.
function answerKeyUp(e)  
{  
	if ( !is_mis && trim(document.getElementById('answer').value) == '' )  {
		document.getElementById('req_val').style.display=  'block'; 
	}
	else  {
		document.getElementById('req_val').style.display= 'none'; 
	}
	if ( e.key === 'Enter' || e.keyCode === 13 ) {  
		nextItem();
	}
}


// validate user answer. par: v => user answer value for current item
function valAnswer()  
{	
	if ( is_mis ) {   // missing values are allowed
	  return true;
	}
	else {   // missing values are not allowed
    let res = ( trim(document.getElementById('answer').value) == '' ) ? false : true;		
		return res; 
  }
}


// https://docs.nagra.com/docs/os/524/maximizing-ui-performance
// https://developer.mozilla.org/en-US/docs/Web/API/Canvas_API/Tutorial/Optimizing_canvas
// http://creativejs.com/resources/requestanimationframe/
// https://developer.mozilla.org/en-US/docs/Web/API/window/requestAnimationFrame
// https://stackoverflow.com/questions/30830126/what-is-difference-between-fillrect0-0-0-1-and-clearrect	
// browser calls this function to update an animation before the next repaint (callback)
// this is 1. callback. this function prepare stimulus in period before stimulus
function prepare(timestamp) 
{
	if (counter == 0)  {   		
		startTime = timestamp;	// reset time when the stimulus sequence started 
    drawString();			// prepare alphanumeric characters => main stimulus
		drawMarks(document.getElementById('can_m'), true);  // prepare marks (cues)
		//document.getElementById('can_s').style.visibility = 'hidden';
		//document.getElementById('can_m').style.visibility = 'hidden';
	}	
	counter= 1;
	let elapsed = timestamp - startTime;  // const elapsed = timestamp - startTime; 					
	if ( elapsed <= before )  {  // wait => var before => Delay time before show each stimulus (ms)
		window.requestAnimationFrame(prepare); // Note: requestAnimationFrame() is 1 shot.
	}
	else  { 		  
		startTime = timestamp;	// timestamp when the alphanumeric characters display started 
		document.getElementById('can_s').style.visibility = 'visible';  // only once
		window.requestAnimationFrame(displayString);  // set another callback => dispatch
	}		
}


// draw all stimulus (alphanumeric characters + delay or parallel marks/cues) 
function drawString() 
{
	let can = document.getElementById('can_s'), ctx = can.getContext('2d'), width = can.width, height = can.height;					
	ctx.fillStyle= back_clr;
	ctx.fillRect(0, 0, width, height);    // ctx.clearRect(0, 0, w, h);
  ctx.fillStyle= pen_clr;
	width /=  col_cnt; // cell width
  height /= row_cnt;  //cell height
	let r = Math.floor( Math.min(width,height)/2 );  
	ctx.font = r.toString() + 'px ' + font_family;   	// update font	
	let cnt= 0, left, top;   // string in each cell array index
	for (let row = 0; row < row_cnt; row++)
	{	   	
		for (let col= 0; col < col_cnt; col++)  {                						
			left= col * width;   // cell left bound
			top = row * height;  // cell top bound      
			ctx.fillText(stim[item][cnt], Math.round(left+width/2), Math.round(top+height/2));
			cnt++;
		}
	} 
	if ( d_time[item] == 0 )  {   
    drawMarks(can, false);
	}		
}	


/*	draw marks (cues) after main stimulus and parallel with main stimulus (if marks delay time == 0)
		can => canvas element: document.getElementById('can_m') or document.getElementById('can_s')
		is_clear => if true : repaint (use fillStyle), else only draw marks */
function drawMarks(can, is_clear) 
{
	let width = can.width, height = can.height, ctx = can.getContext('2d');					
	if ( is_clear )  {
		ctx.clearRect(0, 0, width, height);
		ctx.fillStyle= back_clr;	
	}   
  width /=  col_cnt;  // cell width
  height /= row_cnt;  // cell height
	let r = Math.floor( Math.min(width,height)/2 );  
	ctx.font = r.toString() + 'px ' + font_family;   	// update font		
	let cnt= 0, ind= 0, left, top;  // cnt, ind => index for cell and target  
	for (let row = 0; row < row_cnt; row++)
	{	   	
		for (let col= 0; col < col_cnt; col++)
		{  
      // if target character is in this cell draw marks	
			if ( cnt == target[item][ind] )    // if ( ind < target[item].length && cnt == target[item][ind] ) 
			{   			
				left= col * width;   // cell left bound
				top = row * height;  // cell top bound  
				// draw marks (cues)
				ctx.beginPath();
        // mtype = "1" => stroke rectangle, "2" => stroke circle, "3" => fill rectangle, "4"=> fill circle				
				if ( mtype == '1' || mtype == '3')  {
					ctx.rect(left, top, width, height); 
				}
				else  {
					ctx.arc(Math.round(left+width/2), Math.round(top+height/2), r, 0, 2*Math.PI);   //  //ctx.arc(x,y,r,sAngle,eAngle,counterclockwise);  
				}
				if ( mtype == '1' || mtype == '2')  {
					ctx.stroke();
				}
			  else  {
					ctx.fillStyle= pen_clr;
					ctx.fill();
				}
				ind++;  
			}
			cnt++;
		}
  }
}	


// handle alphanumeric characters duration  
// trigger end of display of alphanumeric characters and set new callback for mark   
function displayString(timestamp)  {
	let elapsed = timestamp - startTime; 
	if ( elapsed < s_time[item] ) {  //   s_time => alphanumeric characters duration (ms) for this item
		window.requestAnimationFrame(displayString);  // wait
	}
	else  {    	
		document.getElementById('can_s').style.visibility = 'hidden';  // only once  
		startTime = timestamp;	// timestamp when the alphanumeric characters display ended
		window.requestAnimationFrame(displayDelay);  // set another callback
	}
}


// handle delay period between end of alphanumeric characters and begin of rectangle cue (mark)
// trigger begin of mark display
function displayDelay(timestamp)  {
	let elapsed = timestamp - startTime; 
	if ( elapsed < d_time[item] )  {   // 	 d_time => variable Mark delay time (ms) for this item
		window.requestAnimationFrame(displayDelay);  // wait
	}
	else  {   	
		document.getElementById('can_m').style.visibility = 'visible';  // only once  
		startTime = timestamp;	// timestamp when the mark (cue) display start
		window.requestAnimationFrame(displayMark);  // set another callback - last
	}
}


// handle mark duration  
// trigger end of mark display and handle after period => setTimeOut() for showItem() 
function displayMark(timestamp)  {
	let elapsed = timestamp - startTime; 
	if ( elapsed < m_time[item] )  {   // 	m_time =>  Mark duration time (ms) for this item
		window.requestAnimationFrame(displayMark);  // wait
	}
	else  {     	
		document.getElementById('can_m').style.visibility = 'hidden';  // only once	
		setTimeout(showItem, after);  // this time period is not precize
	}
}


// return random integer in range [min..max]
function getRndInt(min, max) {
		return Math.floor(Math.random() * (max - min + 1) ) + min;
}


// trigger delayed stimulus draw (animation frame)
function delayDraw()
{
	repeated[item]++;
	counter= 0;
	window.requestAnimationFrame(prepare);	// activate 1. callback Window.requestAnimationFrame()  	
}	


// show current item_div panel/div. items are in range: 0..item_cnt-1     
function showItem() 
{   		
  if ( repeated[item] == 0 )  {  // only first time when item is displayed
		document.getElementById('item_div').style.display= 'none';     
		setTimeout(delayDraw, before);  // this time period is not precize 
	} 
	else 
	{ 
		document.getElementById('item_div').style.display= 'block';  		
		document.getElementById('answer').value= ans[item];	   // user answer		
		document.getElementById('answer').focus();				
		if (demo)  
		{
      let len = target[item].length, tmp= '';			
			for (let j= 0; j< len; j++) {
			  tmp+= stim[item][target[item][j]];  // convert index(s) into text	
			}
			document.getElementById('demo_sol').textContent= 'Target character(s)= ' + tmp; 
		} 		
		document.getElementById('req_val').style.display= 'none'; // hide required warning 		
		updateFooter();
	}	   
}


// update footer (navigation buttons, progress bar,..)
function updateFooter()
{
	document.getElementById('prior').style.display= (item > 0 && is_back) ? 'inline': 'none';  // set footer navigation prior button (left arrow) 				 
	let btn= document.getElementById('next');     
	if(item == stim.length-1)  {       
		btn.style.backgroundImage=  "url('img/end.png')";  // update "Next item" image on "right arrow" button 
		btn.title= 'End of task';
	}
	else  {
		btn.style.backgroundImage= "url('img/forward.png')"; 
		btn.title= 'Next';
	} 
	if (prg_type)  {    // update Progress bar => show current item count and remaining item count on Progress Bar (default)                                                                                                                                                       	
		let percent = Math.round(100*(item_cnt-(item+1))/item_cnt);      
		document.getElementById('progress').setAttribute('style', 'width:'+percent+'%');    // !important
		document.getElementById('nav_items').textContent= (item + 1) + ' / ' + item_cnt;   // item/total
	}	
}	


/* user => user entered value for this item. 
   correct => correct answer (whole content) for this item => joined characters as string
   this function is called for each item
   return number of points => For each stimulus: 1 point is added for each correctly entered alphanumeric character */
function getScore(user, correct)
{
	let tmp= user;  // copy of user answer for this item
	let res= 0, up= correct.length, ind;
	for (let i= 0; i < up; i++)  
	{ 
	  ind = tmp.indexOf(correct[i]);
		if ( ind != -1 ) {     //  str.indexOf(find, str.length) !== -1;
			res++;
			tmp= remove(tmp, ind, 1);  // delete this character (each user char can be scored only once)
		}
	}
	return res;
}


// http://stackoverflow.com/questions/10398931/javascript-how-to-remove-text-from-a-string
function remove(str, startIndex, count) {
    return str.substr(0, startIndex) + str.substr(startIndex + count);
}


// end of task (research)
// if (is_error== true) => page (window) was resized during task (when is_task==true). Only show html error warning
function theEnd(is_error)  
{     
	is_task= false;    
	
	document.getElementById('task_section').style.display= 'none';	
  document.getElementById('end').style.display= 'block';   
 
	if (is_error) {
		document.getElementById('error').style.display= 'block';
		document.getElementById('result_wrapper').style.display= 'none';
		document.getElementById('score').textContent= '';
		return;
	}
	
  document.getElementById('error').style.display= 'none';
	document.getElementById('result_wrapper').style.display= 'block';
	
	let e= document.getElementById('variable');  
	let var_name = e.options[e.selectedIndex].textContent;
  let	header =  ['Item', var_name, 'Correct', 'Entered value', 'Points'];	 
  // my_var => array of all variable values 
	// var_g => variable group by var_v value   	
	let mis= 0, sum= 0, rows = [], points= [], my_var = [], var_g= {};
	switch (variable)  {				
		case '1':				
			my_var= s_time;
			break;
		case '2':				
			my_var= d_time;
			break;
		case '3':				
			my_var= m_time;
			break;
	}			 	 
	for (let i = 0; i < item_cnt; i++)  
	{
    ans[i]= trim(ans[i]);
		let is_mis = ans[i].length == 0 ? true : false;
		let cor = '';
		let len= target[i].length;
		for (let j= 0; j< len; j++)  {
		  cor += stim[i][target[i][j]]; //  stim[i][target[i]];  // convert index into text 
		}
		if ( is_mis )  {  
		  mis++;		// mis = number of missing items in task 			
    }		
		else  
		{  	  
			if ( alfa_type == '2' ) {  		//  <option value="2">Letters-lowercase</option>
				ans[i]= ans[i].toLocaleLowerCase();
			}
			else if ( alfa_type == '3' )  {  // <option value="3">Letters-UPPERCASE</option>
				ans[i]= ans[i].toLocaleUpperCase();
			}       			
		}		
    let point= is_mis ? 0 : getScore(ans[i], cor);
		points.push(point);
    sum+= point;		// sum of points			
		rows[i]= [ parseInt(i+1,10), my_var[i], cor, ans[i], point ];
		
		// group by variable value. for future decriptive statistics (table2)
		let key = my_var[i];   // key => last one pushed into var_v  
		if ( typeof var_g[key] === 'undefined' )  {
			var_g[key] = [];
		}
		var_g[key].push( point );
	}		
  rows.push( ['Sum', '', '', '', sum ] );	
	createTableArray(document.getElementById('table1'), rows, header);	
	
	// calculate decriptive statistics for the points => table2 => group by variable values 	  
	const D = 3;  // number of decimals
	rows = [];   // reset table rows (for table2)		 
	header=  ['Group/Variable', 'Average', 'Standard Deviation', 'N']; // reset header	
	let arr= [], n, m, sd, var_val= [], m_var= [], sd_var=[];
	rows = [];  // reset table rows (for table2)
	for (let key in var_g)  {
		arr = var_g[key];   //  grouped by for this key (variable value)
		n = arr.length;
		m =   n > 0  ? mean(arr).toFixed(D) : '';     // mean only for this variable value (group by)
		sd =  n > 1  ? stDev(arr, m).toFixed(D): '';  // standard deviation only for this variable value (group by) 	 	
		rows.push( [key, m, sd, n] );
    m_var.push(m);
    sd_var.push(sd);
    var_val.push(key);		
	}
		
	n =  points.length;  // number of all items in task
	m =  n > 0  ? mean(points).toFixed(D) : '';      // mean for ALL items in task
	sd = n > 1  ? stDev(points, m).toFixed(D): '';  // standard deviation for ALL items in task 
	rows.push( ['All items', m, sd, n] );  // last row
	
	createTableArray(document.getElementById('table2'), rows, header);  
		
	// update 'score' label => above "Repeat" button
	document.getElementById('score').textContent= 'Sum of points= ' + sum + '. '; 
	switch (n)  {
		case 0: 
			document.getElementById('score').textContent+= 'Descriptive statistics cannot be calculated. Missing items= ' + mis;
			break;
		case 1: 
			document.getElementById('score').textContent+= 'Average points= ' + m + '  Standard deviation cannot be calculated. Missing items= ' + mis;
			break;	
		default:
			document.getElementById('score').textContent+= 'Average points= ' + m + '  Standard Deviation= ' + sd + '. Missing items= ' + mis; 
	}
	
	document.getElementById('l_ts').textContent='Last completed research: ' + new Date();
  document.getElementById('l_res').style.display= 'block';
		
	try 
  {
    if (store && is_storage) {
			setStorage(sum, mis, m, sd, var_name, var_val.join(' '), m_var.join(' '), sd_var.join(' '));  
    }
  }        
  catch (e)  { 
    document.getElementById('table_grp').style.display= 'block';
		document.getElementById('table_grp').textContent= "Your browser probably doesn't have support for the Web Storage Object or access to the Storage Object is disabled. Check your browser settings. In some browsers (IE11, Edge) Storage Object is not enabled when accessed in a local file (offline)."; 
  }  
 
}


// return arithmetic mean (mean, average) of input array (arr)
// no validation => input array must be in valid number format	
function mean(arr)  
{	
	let len= arr.length, sum= 0;
	for (let i=0; i<len; i++) {
		sum+= arr[i];
	}
	return (sum/arr.length);
}	


// return Sample Standard Deviation (with Bessel's correction) => N-1
// no validation => input array must be in valid number format
// arr = array argument;  mean = average for input arr 
function stDev(arr, mean) 
{
	let len= arr.length, sum= 0, res;
	for (let i=0; i<len; i++) {
		sum+= Math.pow(arr[i]-mean, 2);	 // Math.pow(base, exponent)
	}
	res= Math.pow(sum/(len-1), 0.5);   // denominator = N-1
	return res;
}

 
/* Store the results of the last completed task in the user's browser
   set user data into window.localStorage. Parameters:	 
		 points = sum of points in task
		 mis = number of missing items in task
		 m = mean for all items in task
		 sd = standard deviation for all items in task    
		 var_name = selected variable name
		 var_val = variable values (group by)
		 m_var = means for all different variable values (group by)
		 sd_var = standard deviatins for all different variable values (group by variable values)  */
function setStorage(points, mis, m, sd, var_name, var_val, m_var, sd_var)   
{   
	let oldItems = JSON.parse(localStorage.getItem('vprt-v1.0')) || []; 
  let id= oldItems.length+1;     	
	// set new "record"
  let newItem = {
      'id' : id,
      'date' : setDate(),   //new Date(),  		        
			'points' : points, 
			'missing' : mis,
			'mean' : m,
			'sd' : sd,
			'variable' : var_name,			
			'mean-var' : m_var,
			'sd-var' : sd_var,			
			'FPS' : FPS,
			'items': item_cnt,			
			'string-time' : (variable == "1") ? var_val : s_time[0],
			'mark-delay ' : (variable == "2") ? var_val : d_time[0],		
			'mark-time'   : (variable == "3") ? var_val : m_time[0],				
			'time-before': before,
			'time-after' : after,
			'alpha-type': alfa_type,
			'char-memo': target_cnt,
			'rows': row_cnt,
			'columns': col_cnt,
			'mark-type' : mtype,
			'is-back' : is_back ? 1 : 0,
			'is-missing' : is_mis ? 1 : 0, 	
			'demo': demo ? 1 : 0		
  };

  oldItems.push(newItem);

  localStorage.setItem('vprt-v1.0', JSON.stringify(oldItems));
  
	document.getElementById('table_grp').style.display= 'block';
  document.getElementById('store_id').style.display= 'block';
  document.getElementById('store_id').textContent= 'The results are stored in your browser. ID= ' + id ;
  
	createTableJSON(document.getElementById('table_grp'), oldItems);  
  
}


// Formats the date in the desired format. If you do not want this string format do not call this function. use: new Date()
function setDate()  
{
	let today = new Date();
	let dd = today.getDate();
	let mm = today.getMonth()+1; 
	let yyyy = today.getFullYear();
	//if(dd<10) { dd='0'+dd; } 
	if(mm<10) {	mm='0'+mm; } 
	today = dd+'.' + mm + '.' +yyyy;
	return today;
}


// convert JSON data dynamically to HTML Table using JavaScript - Dynamically Create table with rows in JavaScript
// downloaded and adapted from source: https://www.encodedna.com/javascript/populate-json-data-to-html-table-using-javascript.htm
function createTableJSON(elem, my_json) 
{  
  // EXTRACT VALUE FOR HTML HEADER. 
  let col = [];
  for (let i = 0; i < my_json.length; i++) {
    for (let key in my_json[i]) {
        if (col.indexOf(key) === -1) {
            col.push(key);
        }
    }
  } 
  let table = document.createElement("table");
  // CREATE HTML TABLE HEADER ROW USING THE EXTRACTED HEADERS ABOVE.
  let tr = table.insertRow(-1);                   // TABLE ROW.
  for (let i = 0; i < col.length; i++) {
      let th = document.createElement("th");      // TABLE HEADER.
      th.innerHTML = col[i];
      tr.appendChild(th);
  }
  // ADD JSON DATA TO THE TABLE AS ROWS.
  for (let i = 0; i < my_json.length; i++) 
  {
    tr = table.insertRow(-1);
    for (let j = 0; j < col.length; j++) {
      let tabCell = tr.insertCell(-1);
      tabCell.innerHTML = my_json[i][col[j]];
    }
  }
  // FINALLY ADD THE NEWLY CREATED TABLE WITH JSON DATA TO A CONTAINER.
  elem.innerHTML= "";
  elem.appendChild(table);
}


/* convert array data dynamically to HTML table - dynamically create html table
   elem == DOM html element (node)
   data => 1-dim or 2-dim array (matrix)
   col => string column descriptions. Set an empty string if you want to skip a column inside a 2-dim array	 */ 
function createTableArray(elem, data, col) 
{  
	let table= document.createElement('table');  // html table
  let tr= table.insertRow(-1), th, cell, tmp;  // table row, table header, table cell 	
  // create html table header row using the extracted headers above  
  for (let i in col) {
		if ( col[i] == '' ) { continue; } // set an empty string if you want to skip a column inside a 2-dim array
		th= document.createElement('th');      
		th.innerHTML = col[i];
		tr.appendChild(th);
  }
  if ( is2dArray(data) ) 
	{ // add 2-dim array data to the table as rows		
		tmp= data.length;
		for (let i = 0; i < tmp; i++) 
		{
			tr= table.insertRow(-1);
			for (let j in col) {
				if (col[j] == '') { continue; }  // set an empty string if you want to skip a column inside a 2-dim array
				cell= tr.insertCell(-1);
				cell.innerHTML = data[i][j];
			}
		}
  }
	else 
	{	// add 1-dim array data to the table as only one row
		tr= table.insertRow(-1);
		tmp=data.length;
		for (let i = 0; i < tmp; i++)	{
			cell= tr.insertCell(-1);
			cell.innerHTML = data[i];
		}
	}	
  elem.innerHTML= '';
  elem.appendChild(table); // finally add the newly created table with string data to a html container
}


// test if array is 2-dim (return true) or not (return false)
function is2dArray(arr)
{
	if(arr[0] === undefined)	{
		return false;
	}
	else	{
		return (arr[0].constructor === Array);
	}
}


// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
// save to window global vars all user Options/Settings
function hideOption()  
{
  setVarPar();   
  document.getElementById('options').style.display = 'none'; 
  document.getElementById('start').setAttribute('class', 'center-xy');   // ex 'child' class 
  document.getElementById('show_option').style.display = 'block';   
  document.getElementById('start_div').style.display = 'block';        //.removeAttribute("disabled");  
  document.getElementById('hide_option').style.display = 'none';   
  document.getElementById('instructions').style.display= 'block';
  refMsg(); 
	validate();
  setIns();	
}

// begin of new code for export localStorage app data:  data= localStorage.getItem('myApp-v1.0');

// show form with application options
// <!-- IMPORTANT: showOption() and hideOption() buttons must be outside form element -->
function showOption()   
{
  document.getElementById('hide_option').style.display= 'block';
  document.getElementById('options').style.display= 'block';
  document.getElementById('start').classList.remove('center-xy');  // ex 'child' class    
  document.getElementById('start_div').style.display= 'none';    
  document.getElementById('show_option').style.display= 'none';  
  document.getElementById('instructions').style.display= 'none';     
  document.getElementById('store_oper').style.display= 'none';
	if (document.getElementById('store_group').disabled === false && localStorage.getItem('vprt-v1.0')) 
  {
    let cnt= JSON.parse(localStorage.getItem('vprt-v1.0')).length;                                                        
    document.getElementById('store_cnt').textContent=  'You have a total of ' + cnt  + ' results stored in your web browser.';
    if (cnt > 0) {
			document.getElementById('store_oper').style.display= 'table';   // xxx
		}
	}
  dirty= false;
  document.getElementById('is_valid').textContent='';  // tmp reset validator message	
}


// clear web browser storage data for this application
function clrStorage()  
{  
  let cnt= localStorage.getItem('vprt-v1.0') ? JSON.parse(localStorage.getItem('vprt-v1.0')).length : 0;                                                        
  if (cnt === 0)  {
    //alert('You have no data to delete');
    return;
  }  
  if ( !window.confirm('You have a total of ' + cnt  + ' results stored in your web browser. Continue?') ) { 
   return;
 }                                             
 if (window.confirm('Delete total saved application results: '  + cnt + '. This action cannot be undone. Final confirmation?') ) 
 {  
   window.localStorage.removeItem('vprt-v1.0'); 
   document.getElementById('store_cnt').textContent= 'All data has been deleted. No data.';
	 document.getElementById('store_oper').style.display= 'none';
 }  
}


// copy data from user web browser to the Clipboard 
function copyStorage()  
{  
  let data= localStorage.getItem('vprt-v1.0');
	if (!data)  {
    //alert('You have no data to Copy');
    return;
  }  
	data = JSON.parse(data); 
	let arr= jsonToArray(data);	
	copy2DToClipboard(arr);
}	


// convert json object into 2D array
// data = JSON object
// return 2D array
function jsonToArray(data)  
{
   // extract column (fields) names  
  let total= data.length, col = [], row, res=[];  // res= result (2d array) 
  for (let i = 0; i < total; i++) {
    for (let key in data[i]) {
			if (col.indexOf(key) === -1)  {
				col.push(key);  // col= columns (fields) names
			}
    }
  } 
	res.push(col);  // add field names as 1 row
  let c = col.length;    // column count  
  // add JSON data to the 2d array as rows.
  for (let i = 0; i < total; i++) 
  {
		row= [];
    for (let j = 0; j < c; j++)  {     
      row.push(data[i][col[j]]);
    }
		res.push(row);
  }
	return res;	
}


// copy 2D array to Clipboard
function copy2DToClipboard(arr) {
  let csv = '', row, cell;
  for (row = 0; row < arr.length; row++) {
    for (cell = 0; cell < arr[row].length; cell++) {
      csv += (arr[row][cell]+'').replace(/[\n\t]+/g, ' ');
      if (cell+1 < arr[row].length) csv += '\t';
    }
    if (row+1 < arr.length) { 
			csv += '\n'; 
		}
  }
  copyTextToClipboard(csv);
}


// copy 2D array (or text) to Clipboard
function copyTextToClipboard(txt) 
{	
	if (!navigator.clipboard) {
    fallbackCopy(txt);
    return;
  }
  navigator.clipboard.writeText(txt).then(function() {
   //alert('Async: Copying to Clipboard was successful!');
  }, function(err) {
    alert('Oops, could not copy data. Error: ' + err);
  });
}


// fallback copy text to clipboard
function fallbackCopy(txt) 
{
  let ta = document.createElement('textarea');
  ta.value = txt; 
  // Avoid scrolling to bottom
  ta.style.top = '0';
  ta.style.left = '0';
  ta.style.position = 'fixed';
  document.body.appendChild(ta);
  ta.focus();
  ta.select();
  try {
    let successful = document.execCommand('copy');
    let msg = successful ? 'successful' : 'unsuccessful';
  } catch (err) {
    alert('Oops, unable to copy. Error: ' + err);
  }
  document.body.removeChild(ta);
}


// displays user saved results (Window.localStorage application data) in new empty web page (tab) 
function newWin()
{ 
	let data = JSON.parse(localStorage.getItem('vprt-v1.0'))
	createTableJSON(document.getElementById('table_grp'), data);
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= document.getElementById('table_grp').innerHTML;	
	let win = window.open('');
  win.document.write('<html><head><title>Saved data</title>' + stl + '</head><body>' + tbl + '</body></html>');
  win.focus();
}


// displays Latest research results in new empty web page (tab) 
function newWinLast()
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex;
	ind++;
	let e = document.getElementById('table' + ind);	
	let stl= '<style>table{margin:1em auto;border:thick solid #000;border-collapse:collapse}td,th{border:medium solid #000;text-align:center;padding:.5em}th{font-weight:bold}</style>';
	let tbl= e.innerHTML; 	
	let win = window.open('');
	win.document.write('<html><head><title>Latest research data</title>' + stl + '</head><body>' + tbl + '</body></html>');
	win.focus();
}


// copy only latest task results (data) to the Clipboard 
function copyLast()  
{ 
  let ind = document.getElementById('last_data_sel').selectedIndex, tbl;
  switch (ind) {
    case 0:
			tbl = document.getElementById('table1').getElementsByTagName('table')[0];  // note: id='table1' is html div element
			break;		
    case 1:
			tbl = document.getElementById('table2').getElementsByTagName('table')[0];  // note: id='table2' is html div element
			break;
	}		
  if (tbl)  {
		let data= table2DArray(tbl);
		if (data.length > 0) {
			copy2DToClipboard(data);
		}
	}
}	


// input argument tbl == html/DOM  table  object
// return 2D array
function table2DArray(tbl) 
{
	let res = [];	 
	for (let i = 0, row; row = tbl.rows[i]; i++) 
	{   //iterate through rows. rows would be accessed using the "row" variable assigned in the for loop
	  res[i]= [];		
		for (let j = 0, col; col = row.cells[j]; j++) {  //iterate through columns. columns would be accessed using the "col" variable assigned in the for loop			 
		  res[i][j]= col.textContent;
		}  
	}
	return res;				
}
